from typing import cast
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl
import litellm
from litellm import CustomStreamWrapper
from litellm.types.utils import StreamingChoices

import langwatch


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    langwatch.get_current_trace().autotrack_litellm_calls(litellm)

    msg = cl.Message(
        content="",
    )

    response = litellm.completion(
        model="groq/llama3-70b-8192",
        messages=[
            {
                "role": "system",
                "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
            },
            {"role": "user", "content": message.content},
        ],
        stream=True,
        stream_options={"include_usage": True},
    )

    for part in cast(CustomStreamWrapper, response):
        if token := cast(StreamingChoices, part.choices[0]).delta.content or "":
            await msg.stream_token(token)

    await msg.update()



import json
from langchain_core.messages import HumanMessage
from langwatch.utils import SerializableWithStringFallback


def test_it_serializes_things():
    serialized = json.loads(
        json.dumps(
            {"messages": [HumanMessage(content="Hello")]},
            cls=SerializableWithStringFallback,
        )
    )

    assert serialized == {"messages": [{"content": "Hello", "role": "user"}]}



from dotenv import load_dotenv
from fastapi.testclient import TestClient

load_dotenv()

import os
from fastapi import FastAPI
from openai import OpenAI
from pydantic import BaseModel
from opentelemetry import trace
from opentelemetry.sdk.resources import Resource
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

client = OpenAI()


# Set up OpenTelemetry trace provider with LangWatch as the endpoint
tracer_provider = trace_sdk.TracerProvider(
    resource=Resource(attributes={"service.name": "fastapi_sample_endpoint"})
)
tracer_provider.add_span_processor(
    SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=f"{os.environ.get('LANGWATCH_ENDPOINT', 'https://app.langwatch.ai')}/api/otel/v1/traces",
            headers={"Authorization": "Bearer " + os.environ["LANGWATCH_API_KEY"]},
        )
    )
)
# Optionally, you can also print the spans to the console.
tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))


app = FastAPI()
# Instrument FastAPI with OpenTelemetry
FastAPIInstrumentor.instrument_app(app)
trace.set_tracer_provider(tracer_provider)
tracer = trace.get_tracer(__name__)


class EndpointParams(BaseModel):
    input: str


@app.post("/")
def fastapi_sample_endpoint(params: EndpointParams):
    with tracer.start_as_current_span("fastapi_sample_endpoint"):
        completion = client.chat.completions.create(
            model="gpt-4o",
            messages=[
                {
                    "role": "system",
                    "content": "You are a helpful assistant that only reply in short tweet-like responses, using lots of emojis.",
                },
                {"role": "user", "content": params.input},
            ],
        )

        return completion.choices[0].message.content


def call_fastapi_sample_endpoint(input: str) -> str:
    test_client = TestClient(app)
    response = test_client.post("/", json={"input": input})

    return response.json()


if __name__ == "__main__":
    import uvicorn
    import os

    # Test one llm call before starting the server
    print(call_fastapi_sample_endpoint("Hello, world!"))

    port = int(os.environ.get("PORT", 9000))
    uvicorn.run(app, host="0.0.0.0", port=port)



import os

os.environ["CHAINLIT_AUTH_SECRET"] = (
    "sSgdH4IfqM/%5swBnuwNdZ8rS/VBDW-WiBF?>DW_YPPKm9p9CiO6PPm6Z4Ih7IL"
)

import datetime
import os.path
import pickle
from typing import Dict, List, Literal, Optional
from dotenv import load_dotenv

from langwatch.types import RAGChunk

load_dotenv()

import chainlit as cl
from langchain.prompts import ChatPromptTemplate
from langchain.schema import HumanMessage
from langchain.schema.runnable.config import RunnableConfig

import langwatch

from langchain_community.document_loaders import WebBaseLoader
from langchain_community.vectorstores.faiss import FAISS
from langchain_openai import ChatOpenAI, OpenAIEmbeddings
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain.tools.retriever import create_retriever_tool
from langchain.agents import AgentExecutor, create_tool_calling_agent
from langchain.tools import BaseTool, StructuredTool, tool
import chainlit as cl
import chainlit.data as cl_data
from chainlit.data.utils import queue_until_user_message
from chainlit.element import Element, ElementDict
from chainlit.socket import persist_user_session
from chainlit.step import StepDict
from chainlit.types import (
    Feedback,
    PageInfo,
    PaginatedResponse,
    Pagination,
    ThreadDict,
    ThreadFilter,
)
from langgraph.graph import StateGraph, START, END, MessagesState
from langgraph.prebuilt import ToolNode
from langgraph.graph.state import CompiledStateGraph

loader = WebBaseLoader("https://docs.langwatch.ai")
docs = loader.load()
documents = RecursiveCharacterTextSplitter(
    chunk_size=1000, chunk_overlap=200
).split_documents(docs)
vector = FAISS.from_documents(documents, OpenAIEmbeddings())
retriever = vector.as_retriever()


now = datetime.datetime.now()

thread_history = [
    {
        "id": "test1",
        "name": "thread 1",
        "createdAt": now,
        "userId": "test",
        "userIdentifier": "admin",
        "steps": [
            {
                "id": "test1",
                "name": "test",
                "createdAt": now,
                "type": "user_message",
                "output": "Message 1",
            },
            {
                "id": "test2",
                "name": "test",
                "createdAt": now,
                "type": "assistant_message",
                "output": "Message 2",
            },
        ],
    },
    {
        "id": "test2",
        "createdAt": now,
        "userId": "test",
        "userIdentifier": "admin",
        "name": "thread 2",
        "steps": [
            {
                "id": "test3",
                "createdAt": now,
                "name": "test",
                "type": "user_message",
                "output": "Message 3",
            },
            {
                "id": "test4",
                "createdAt": now,
                "name": "test",
                "type": "assistant_message",
                "output": "Message 4",
            },
        ],
    },
]
deleted_thread_ids = []

THREAD_HISTORY_PICKLE_PATH = os.getenv("THREAD_HISTORY_PICKLE_PATH")
if THREAD_HISTORY_PICKLE_PATH and os.path.exists(THREAD_HISTORY_PICKLE_PATH):
    with open(THREAD_HISTORY_PICKLE_PATH, "rb") as f:
        thread_history = pickle.load(f)


async def save_thread_history():
    if THREAD_HISTORY_PICKLE_PATH:
        # Force saving of thread history for reload when server restarts
        await persist_user_session(
            cl.context.session.thread_id, cl.context.session.to_persistable()
        )

        with open(THREAD_HISTORY_PICKLE_PATH, "wb") as out_file:
            pickle.dump(thread_history, out_file)


class TestDataLayer(cl_data.base.BaseDataLayer):
    async def get_user(self, identifier: str):
        return cl.PersistedUser(
            id="test", createdAt=now.isoformat(), identifier=identifier
        )

    async def create_user(self, user: cl.User):
        return cl.PersistedUser(
            id="test", createdAt=now.isoformat(), identifier=user.identifier
        )

    async def update_thread(
        self,
        thread_id: str,
        name: Optional[str] = None,
        user_id: Optional[str] = None,
        metadata: Optional[Dict] = None,
        tags: Optional[List[str]] = None,
    ):
        thread = next((t for t in thread_history if t["id"] == thread_id), None)
        if thread:
            if name:
                thread["name"] = name
            if metadata:
                thread["metadata"] = metadata
            if tags:
                thread["tags"] = tags
        else:
            thread_history.append(
                {
                    "id": thread_id,
                    "name": name,
                    "metadata": metadata,
                    "tags": tags,
                    "createdAt": now.isoformat(),
                    "userId": user_id,
                    "userIdentifier": "admin",
                    "steps": [],
                }
            )

    @cl_data.queue_until_user_message()
    async def create_step(self, step_dict: StepDict):
        thread = next(
            (t for t in thread_history if t["id"] == step_dict.get("threadId")), None
        )
        if thread:
            thread["steps"].append(step_dict)

    async def get_thread_author(self, thread_id: str):
        return "admin"

    async def list_threads(
        self, pagination: Pagination, filters: ThreadFilter
    ) -> PaginatedResponse[ThreadDict]:
        return PaginatedResponse(
            data=[t for t in thread_history if t["id"] not in deleted_thread_ids],
            pageInfo=PageInfo(hasNextPage=False, startCursor=None, endCursor=None),
        )  # type: ignore

    async def get_thread(self, thread_id: str):
        thread = next((t for t in thread_history if t["id"] == thread_id), None)
        if not thread:
            return None
        thread["steps"] = sorted(thread["steps"], key=lambda x: x["createdAt"])
        return thread

    async def delete_thread(self, thread_id: str):
        deleted_thread_ids.append(thread_id)

    async def delete_feedback(
        self,
        feedback_id: str,
    ) -> bool:
        return True

    async def upsert_feedback(
        self,
        feedback: Feedback,
    ) -> str:
        return ""

    @queue_until_user_message()
    async def create_element(self, element: "Element"):
        pass

    async def get_element(
        self, thread_id: str, element_id: str
    ) -> Optional["ElementDict"]:
        pass

    @queue_until_user_message()
    async def delete_element(self, element_id: str, thread_id: Optional[str] = None):
        pass

    @queue_until_user_message()
    async def update_step(self, step_dict: "StepDict"):
        pass

    @queue_until_user_message()
    async def delete_step(self, step_id: str):
        pass

    async def build_debug_url(self) -> str:
        return ""


cl_data._data_layer = TestDataLayer()


retriever_tool = create_retriever_tool(
    langwatch.langchain.capture_rag_from_retriever(
        retriever,
        lambda document: RAGChunk(
            document_id=document.metadata["source"], content=document.page_content
        ),
    ),
    "langwatch_search",
    "Search for information about LangWatch. For any questions about LangWatch, use this tool if you didn't already",
)

tools = [retriever_tool]
tool_node = ToolNode(tools)
model = ChatOpenAI(streaming=True).bind_tools(tools)


def should_continue(state: MessagesState):
    messages = state["messages"]
    last_message = messages[-1]
    # If the LLM makes a tool call, then we route to the "tools" node
    if last_message.tool_calls:  # type: ignore
        return "tools"
    return END


def call_model(state):
    messages = state["messages"]
    response = model.invoke(messages)
    # We return a list, because this will get added to the existing list
    return {"messages": [response]}


# Define the graph
workflow = StateGraph(MessagesState)

workflow.add_node("agent", call_model)
workflow.add_node("tools", tool_node)

workflow.add_edge(START, "agent")
workflow.add_conditional_edges("agent", should_continue)
workflow.add_edge("tools", "agent")

# Compile the graph
app = workflow.compile()


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    msg = cl.Message(content="")

    langwatch.get_current_trace().update(
        metadata={
            "user_id": getattr(cl.user_session.get("user"), "identifier", "unknown"),
            "thread_id": cl.context.session.thread_id,
        }
    )

    async for event in app.astream(
        {"messages": [HumanMessage(content=message.content)]},
        config=RunnableConfig(
            callbacks=[
                cl.LangchainCallbackHandler(),
                langwatch.get_current_trace().get_langchain_callback(),
            ]
        ),
    ):
        if "agent" in event:
            if "messages" in event["agent"]:
                for message in event["agent"]["messages"]:
                    await msg.stream_token(message.content)

    await msg.send()

    await save_thread_history()


@cl.password_auth_callback
async def auth_callback(username: str, password: str) -> Optional[cl.User]:
    if (username, password) == ("admin", "admin"):
        return cl.User(identifier="admin")
    else:
        return None


@cl.on_chat_resume
async def on_chat_resume(thread: ThreadDict):
    await cl.Message(f"Welcome back to {thread['name']}").send()
    if "metadata" in thread and thread["metadata"]:
        await cl.Message(thread["metadata"], author="metadata", language="json").send()
    if "tags" in thread and thread["tags"]:
        await cl.Message(
            ",".join(thread["tags"]), author="tags", language="json"
        ).send()



import logging
import langwatch


def get_logger():
    logger = logging.getLogger("LangWatch")
    logger.propagate = False

    if langwatch.debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    if logger.handlers:
        logger.handlers.clear()

    if not logger.handlers:
        console_handler = logging.StreamHandler()
        formatter = logging.Formatter(
            "%(asctime)s - [%(name)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
        )
        console_handler.setFormatter(formatter)
        logger.addHandler(console_handler)

    return logger



import json
import os
import importlib
import sys
from typing import Optional, cast
import pytest
import asyncio
import chainlit as cl

from chainlit.context import init_http_context
from langwatch.tracer import ContextTrace
from opentelemetry.sdk.trace.export import (
    SpanExportResult,
    SpanExporter,
    SimpleSpanProcessor,
)


last_trace: Optional[ContextTrace] = None
trace_urls: dict[str, str] = {}


original_init = ContextTrace.__init__


def patched_init(self, *args, **kwargs):
    global last_trace
    last_trace = self
    self._force_sync = True

    return original_init(self, *args, **kwargs)


ContextTrace.__init__ = patched_init


class TraceIdCapturerExporter(SpanExporter):
    def export(self, spans):
        global last_trace

        context = spans[0].get_span_context()
        if context is not None:
            trace_id = context.trace_id.to_bytes(16, "big").hex()
            last_trace = ContextTrace(trace_id=trace_id)
        return SpanExportResult.SUCCESS


def get_example_files():
    examples_dir = os.path.join(os.path.dirname(__file__), "..", "examples")
    opentelemetry_dir = os.path.join(examples_dir, "opentelemetry")
    return [
        f
        for f in os.listdir(examples_dir)
        if f.endswith(".py") and not f.startswith("__")
    ] + [
        "opentelemetry/" + f for f in os.listdir(opentelemetry_dir) if f.endswith(".py")
    ]


@pytest.mark.parametrize("example_file", get_example_files())
@pytest.mark.asyncio
async def test_example(example_file):
    if example_file == "batch_evalutation.py":
        pytest.skip("batch_evalutation.py is not a runnable example")
    if example_file == "opentelemetry/openllmetry_anthropic_bot.py":
        pytest.skip(
            "openllmetry anthropic has a bug starting another async process inside"
        )
    if example_file == "opentelemetry/openllmetry_openai_bot.py":
        pytest.skip(
            "openllmetry openai has a bug starting another async process inside"
        )
    if example_file == "langchain_rag_bot_vertex_ai.py":
        pytest.skip(
            "langchain_rag_bot_vertex_ai.py is broken due to a bug in current langchain version of global state mutation when running together with other langchain"
        )

    global last_trace
    last_trace = None

    # Dynamically import the main function from the example file
    module_name = (
        f"examples.{example_file[:-3].replace('/', '.')}"  # Remove .py extension
    )
    module = importlib.import_module(module_name)
    init_http_context()

    # Skip if there's no main function
    main_func = getattr(module, "main", None)
    if "fastapi" in example_file:
        main_func = getattr(module, "call_fastapi_sample_endpoint", None)
    if main_func is None:
        pytest.skip(f"No main function found in {example_file}")

    if "opentelemetry" in example_file:
        tracer_provider = getattr(module, "tracer_provider")
        # Remove console printing exporter
        tracer_provider._active_span_processor._span_processors = (
            tracer_provider._active_span_processor._span_processors[0],
        )
        # Capture trace id
        tracer_provider.add_span_processor(
            SimpleSpanProcessor(TraceIdCapturerExporter())
        )

    on_chat_start = getattr(module, "on_chat_start", None)
    if on_chat_start is not None:
        await on_chat_start()

    # Create a mock cl.Message
    content = (
        "who is the oldest person?"
        if "span_evaluation" in example_file
        else "what is LangWatch?" if "rag" in example_file else "hello"
    )
    mock_message = content if "fastapi" in example_file else cl.Message(content=content)

    # Call the main function
    try:
        if asyncio.iscoroutinefunction(main_func):
            await main_func(mock_message)
        else:
            main_func(mock_message)
    except Exception as e:
        if str(e) != "This exception will be captured by LangWatch automatically":
            pytest.fail(f"Error running main function in {example_file}: {str(e)}")

    if last_trace is not None:
        last_trace = cast(ContextTrace, last_trace)
        last_trace.send_spans()
        trace_urls[example_file] = last_trace.share()
        print(json.dumps(trace_urls, indent=2))


@pytest.mark.asyncio
async def test_example_legacy_langchain_pydantic_bot():
    import subprocess

    result = subprocess.run(
        ["poetry", "run", ".venv/bin/python", "legacy_langchain_pydantic_bot.py"],
        cwd="examples/legacy",
        capture_output=True,
        text=True,
        check=False,
    )

    print(result.stdout)

    if result.returncode != 0:
        pytest.fail(f"Failed to run legacy_langchain_pydantic_bot.py: {result.stderr}")



from dotenv import load_dotenv
from langchain_openai import ChatOpenAI

load_dotenv()

import chainlit as cl
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain.schema.runnable import Runnable, RunnableMap
from langchain.schema.runnable.config import RunnableConfig
from langchain.memory import ConversationBufferMemory

import langwatch


session_memories: dict = {}


@cl.on_chat_start
async def on_chat_start():
    memory = session_memories.setdefault(
        cl.user_session.get("session_id"),
        ConversationBufferMemory(return_messages=True, memory_key="chat_history"),
    )
    ingress = RunnableMap(
        {
            "input": lambda x: x["input"],
            "chat_history": lambda x: memory.load_memory_variables(x)["chat_history"],
        }
    )
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", "You are a helpful assistant."),
            MessagesPlaceholder(variable_name="chat_history"),
            ("human", "{input}"),
        ]
    )
    llm = ChatOpenAI(temperature=0.3, model="gpt-4o-mini", max_tokens=4096)
    runnable = ingress | prompt | llm

    cl.user_session.set("runnable", runnable)
    cl.user_session.set("memory", memory)


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    runnable: Runnable = cl.user_session.get("runnable")  # type: ignore
    memory = cl.user_session.get("memory")  # type: ignore

    msg = cl.Message(content="")

    langwatch.get_current_trace().update(
        metadata={"customer_id": "customer_example", "labels": ["v1.0.0"]}
    )

    async for chunk in runnable.astream(
        {"input": message.content},
        config=RunnableConfig(
            callbacks=[
                cl.LangchainCallbackHandler(),
                langwatch.get_current_trace().get_langchain_callback(),
            ]
        ),
    ):
        await msg.stream_token(chunk.content)

    memory.save_context({"input": message.content}, {"output": msg.content})  # type: ignore

    await msg.send()



import time
import uuid
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import langwatch


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    trace = langwatch.trace(
        trace_id=uuid.uuid4(),
    )
    span = trace.span(
        type="span",
        input=message.content,
    )
    nested_span = span.span(
        type="llm", input=message.content, model="openai/gpt-4o-mini"
    )

    time.sleep(1)  # generating the message...
    generated_message = "Hello there! How can I help from low level?"

    nested_span.end(output=generated_message)
    span.end()

    trace.deferred_send_spans()

    await msg.stream_token(generated_message)
    await msg.update()



# This example uses the OpenTelemetry instrumentation for OpenAI from OpenLLMetry: https://pypi.org/project/opentelemetry-instrumentation-openai/

from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import os
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.sdk import trace as trace_sdk
from opentelemetry.sdk.trace.export import ConsoleSpanExporter, SimpleSpanProcessor

from opentelemetry.instrumentation.anthropic import AnthropicInstrumentor
import anthropic

client = anthropic.Anthropic()


# Set up OpenTelemetry trace provider with LangWatch as the endpoint
tracer_provider = trace_sdk.TracerProvider()
tracer_provider.add_span_processor(
    SimpleSpanProcessor(
        OTLPSpanExporter(
            endpoint=f"{os.environ.get('LANGWATCH_ENDPOINT', 'https://app.langwatch.ai')}/api/otel/v1/traces",
            headers={"Authorization": "Bearer " + os.environ["LANGWATCH_API_KEY"]},
        )
    )
)
# Optionally, you can also print the spans to the console.
tracer_provider.add_span_processor(SimpleSpanProcessor(ConsoleSpanExporter()))

AnthropicInstrumentor().instrument(tracer_provider=tracer_provider)


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    completion = client.messages.create(
        model="claude-3-5-sonnet-20240620",
        max_tokens=1000,
        temperature=0,
        stream=True,
        system="You are a world-class poet. Respond only with short poems.",
        messages=[
            {
                "role": "user",
                "content": [{"type": "text", "text": "Why is the ocean salty?"}],
            }
        ],
    )

    for part in completion:
        if part.type == "content_block_delta":
            await msg.stream_token(part.delta.text or "")

    await msg.update()



import time
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import langwatch


@langwatch.span(type="llm")
def generate(message: str):
    time.sleep(1)  # generating the message...

    generated_message = "Hello there! How can I help?"

    langwatch.get_current_span().update(model="custom_model")

    return generated_message


@cl.on_message
@langwatch.trace()
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    langwatch.get_current_trace().update(
        metadata={"user_id": "123", "question_id": "456", "labels": ["some-label"]}
    )

    generated_message = generate(message.content)

    await msg.stream_token(generated_message)
    await msg.update()



import copy
from importlib.metadata import version
import os
from concurrent.futures import Future, ThreadPoolExecutor
from random import random

import time
from types import ModuleType
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
    cast,
)
from uuid import UUID
from warnings import warn
from deprecated import deprecated
import httpx

import nanoid
from langwatch.logger import get_logger
from langwatch.types import (
    BaseSpan,
    ChatMessage,
    Conversation,
    Evaluation,
    EvaluationResult,
    EvaluationTimestamps,
    LLMSpan,
    Money,
    MoneyDict,
    SpanMetrics,
    SpanParams,
    RAGChunk,
    RAGSpan,
    Span,
    SpanInputOutput,
    SpanTimestamps,
    SpanTypes,
    CollectorRESTParams,
    TraceMetadata,
    TypedValueEvaluationResult,
)
from langwatch.utils import (
    SerializableWithStringFallback,
    autoconvert_rag_contexts,
    autoconvert_typed_values,
    capture_exception,
    milliseconds_timestamp,
    reduce_payload_size,
)
from retry import retry

import langwatch
import contextvars  # Import contextvars module

import functools
import inspect

try:
    from openai import (
        OpenAI,
        AsyncOpenAI,
        AzureOpenAI,
        AsyncAzureOpenAI,
    )
except ImportError:
    pass

T = TypeVar("T", bound=Callable[..., Any])


def get_version():
    """Retrieves the version of the current library."""
    try:
        return version("langwatch")
    except Exception:
        return "unknown"


class ContextSpan:
    trace: Optional["ContextTrace"] = None
    context_token: Optional[contextvars.Token[Optional["ContextSpan"]]] = None
    span: Type["ContextSpan"]

    span_id: Union[str, UUID]
    parent: Optional["ContextSpan"] = None
    _parent_from_context: bool = False
    _capture_input: bool = True
    _capture_output: bool = True
    _ignore_missing_trace_warning: bool = False
    name: Optional[str] = None
    type: SpanTypes = "span"
    input: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None
    output: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None
    error: Optional[Exception] = None
    timestamps: SpanTimestamps
    contexts: Optional[Union[List[RAGChunk], List[str]]] = None
    model: Optional[str] = None
    params: Optional[SpanParams] = None
    metrics: Optional[SpanMetrics] = None

    def __init__(
        self,
        trace: Optional["ContextTrace"] = None,
        span_id: Optional[Union[str, UUID]] = None,
        parent: Optional["ContextSpan"] = None,
        capture_input: bool = True,
        capture_output: bool = True,
        name: Optional[str] = None,
        type: SpanTypes = "span",
        input: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        output: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[SpanTimestamps] = None,
        contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
        model: Optional[str] = None,
        params: Optional[SpanParams] = None,
        metrics: Optional[SpanMetrics] = None,
        ignore_missing_trace_warning: bool = False,
    ) -> None:
        self.trace = trace
        self.parent = parent
        self._capture_input = capture_input
        self._capture_output = capture_output
        self._ignore_missing_trace_warning = ignore_missing_trace_warning
        self.timestamps = SpanTimestamps(started_at=milliseconds_timestamp())
        self.update(
            span_id=span_id or f"span_{nanoid.generate()}",
            name=name,
            type=type,
            input=input,
            output=output,
            error=error,
            timestamps=timestamps,
            contexts=contexts,
            model=model,
            params=params,
            metrics=metrics,
        )
        self.span = cast(
            Type["ContextSpan"],
            lambda **kwargs: ContextSpan(trace=self.trace, parent=self, **kwargs),
        )

    def __enter__(self):
        current_trace = current_trace_var.get() or self.trace
        if current_trace:
            if not self.trace:
                self.trace = current_trace

            if self.trace == current_trace:
                if not self.parent and current_trace.get_current_span():
                    self.parent = current_trace.get_current_span()
                    self._parent_from_context = True
                self.context_token = current_trace.set_current_span(self)
        elif not self._ignore_missing_trace_warning:
            warn("No current trace found, some spans will not be sent to LangWatch")

        return self

    def __exit__(
        self, _exc_type=None, exc_value: Optional[Exception] = None, _exc_traceback=None
    ):
        self.end(error=exc_value)

        if self.trace:
            self.trace.reset_current_span(self.context_token)
            self.context_token = None

    def __call__(self, func: T) -> T:
        span_kwargs = {
            "trace": None,
            "span_id": None,
            "parent": None,
            "capture_input": self._capture_input,
            "capture_output": self._capture_output,
            "name": self.name,
            "type": self.type,
            "input": self.input,
            "output": self.output,
            "error": self.error,
            "timestamps": None,
            "contexts": self.contexts,
            "model": self.model,
            "params": self.params,
            "metrics": self.metrics,
        }

        if inspect.isasyncgenfunction(func):

            @functools.wraps(func)
            async def async_gen_wrapper(*args, **kwargs):
                with ContextSpan(**span_kwargs) as span:
                    span._capture_name_and_input(func, *args, **kwargs)
                    items = []
                    async for item in func(*args, **kwargs):
                        items.append(item)
                        yield item

                    output = (
                        "".join(items)
                        if all(isinstance(item, str) for item in items)
                        else items
                    )
                    span._capture_output_and_maybe_name(func, output)

            return async_gen_wrapper  # type: ignore
        elif inspect.isgeneratorfunction(func):

            @functools.wraps(func)
            def sync_gen_wrapper(*args, **kwargs):
                with ContextSpan(**span_kwargs) as span:
                    span._capture_name_and_input(func, *args, **kwargs)
                    items = []
                    for item in func(*args, **kwargs):
                        items.append(item)
                        yield item

                    output = (
                        "".join(items)
                        if all(isinstance(item, str) for item in items)
                        else items
                    )
                    span._capture_output_and_maybe_name(func, output)

            return sync_gen_wrapper  # type: ignore

        elif inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                with ContextSpan(**span_kwargs) as span:
                    span._capture_name_and_input(func, *args, **kwargs)
                    output = await func(*args, **kwargs)
                    span._capture_output_and_maybe_name(func, output)
                    return output

            return async_wrapper  # type: ignore
        else:

            @functools.wraps(func)  # type: ignore
            def sync_wrapper(*args, **kwargs):
                with ContextSpan(**span_kwargs) as span:
                    span._capture_name_and_input(func, *args, **kwargs)
                    output = func(*args, **kwargs)  # type: ignore
                    span._capture_output_and_maybe_name(func, output)
                    return output

            return sync_wrapper  # type: ignore

    def _capture_name_and_input(self, func, *args, **kwargs):
        if self.type != "llm":
            self.name = func.__name__
        if self._capture_input:
            sig = inspect.signature(func)
            parameters = list(sig.parameters.values())

            all_args = {
                str(parameter.name): value for parameter, value in zip(parameters, args)
            }
            # Skip self parameters because it doesn't really help with debugging, becomes just noise
            if (
                "self" in all_args
                and len(all_args) > 0
                and parameters[0].name == "self"
            ):
                self_ = all_args["self"]
                try:
                    self.name = f"{self_.__class__.__name__}.{func.__name__}"
                except:
                    pass
                del all_args["self"]

            if kwargs and len(kwargs) > 0:
                if kwargs:
                    all_args.update(kwargs)

            if len(all_args) == 0:
                return

            self.input = autoconvert_typed_values(all_args)

    def _capture_output_and_maybe_name(self, func, output):
        if not self.output and self._capture_output:
            self.output = autoconvert_typed_values(output)
        if self.type == "llm" and not self.model:
            self.name = func.__name__

    def update(
        self,
        span_id: Optional[Union[str, UUID]] = None,
        name: Optional[str] = None,
        type: Optional[SpanTypes] = None,
        input: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        output: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[SpanTimestamps] = None,
        contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
        model: Optional[str] = None,
        params: Optional[SpanParams] = None,
        metrics: Optional[SpanMetrics] = None,
    ) -> None:
        if span_id:
            self.span_id = span_id
        if name:
            self.name = name
        if type:
            self.type = type
        if input:
            # Avoid late mutations after capturing the value
            self.input = copy.deepcopy(input)
        if output:
            # Avoid late mutations after capturing the value
            self.output = copy.deepcopy(output)
        if error:
            self.error = error
        if timestamps:
            if self.timestamps:
                self.timestamps = {**self.timestamps, **timestamps}
            else:
                self.timestamps = timestamps
        if contexts:
            if self.type == "rag":
                # Avoid late mutations after capturing the value
                self.contexts = copy.deepcopy(contexts)
            else:
                warn(
                    "Trying `contexts` on a non-RAG span, this attribute will be ignored for non-RAG spans, please make sure you set the span type to `rag` by using the decorator as @span(type='rag')"
                )
        if model:
            if self.type == "llm":
                self.model = model
            else:
                warn(
                    "Trying `model` on a non-LLM span, this attribute will be ignored for non-LLM spans, please make sure you set the span type to `llm` by using the decorator as @span(type='llm')"
                )
        if params:
            # Avoid late mutations after capturing the value
            params = copy.deepcopy(params)
            if self.params:
                self.params = {**self.params, **params}
            else:
                self.params = params
        if metrics:
            # Avoid late mutations after capturing the value
            metrics = copy.deepcopy(metrics)
            if self.metrics:
                self.metrics = {**self.metrics, **metrics}
            else:
                self.metrics = metrics

    def add_evaluation(
        self,
        *,
        evaluation_id: Optional[str] = None,
        name: str,
        type: Optional[str] = None,
        is_guardrail: Optional[bool] = None,
        status: Literal["processed", "skipped", "error"] = "processed",
        passed: Optional[bool] = None,
        score: Optional[float] = None,
        label: Optional[str] = None,
        details: Optional[str] = None,
        cost: Optional[Union[Money, MoneyDict, float]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[EvaluationTimestamps] = None,
    ):
        if not self.trace:
            raise ValueError("No trace found, could not add evaluation to span")

        self.trace.add_evaluation(
            evaluation_id=evaluation_id,
            span=self,
            name=name,
            type=type,
            is_guardrail=is_guardrail,
            status=status,
            passed=passed,
            score=score,
            label=label,
            details=details,
            cost=cost,
            error=error,
            timestamps=timestamps,
        )

    def evaluate(
        self,
        slug: str,
        name: Optional[str] = None,
        input: Optional[str] = None,
        output: Optional[str] = None,
        expected_output: Optional[str] = None,
        contexts: Union[List[RAGChunk], List[str]] = [],
        conversation: Conversation = [],
        settings: Optional[dict] = None,
        as_guardrail: bool = False,
    ):
        return langwatch.evaluations.evaluate(
            span=self,
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            conversation=conversation,
            settings=settings,
            as_guardrail=as_guardrail,
        )

    async def async_evaluate(
        self,
        slug: str,
        name: Optional[str] = None,
        input: Optional[str] = None,
        output: Optional[str] = None,
        expected_output: Optional[str] = None,
        contexts: Union[List[RAGChunk], List[str]] = [],
        conversation: Conversation = [],
        settings: Optional[dict] = None,
        as_guardrail: bool = False,
    ):
        return await langwatch.evaluations.async_evaluate(
            span=self,
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            conversation=conversation,
            settings=settings,
            as_guardrail=as_guardrail,
        )

    def end(
        self,
        name: Optional[str] = None,
        type: Optional[SpanTypes] = None,
        input: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        output: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[SpanTimestamps] = None,
        contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
        model: Optional[str] = None,
        params: Optional[SpanParams] = None,
        metrics: Optional[SpanMetrics] = None,
    ):
        finished_at = milliseconds_timestamp()

        self.update(
            name=name,
            type=type,
            input=input,
            output=output,
            error=error,
            timestamps={
                "finished_at": finished_at,
                **(timestamps or {}),
            },
            contexts=contexts,
            model=model,
            params=params,
            metrics=metrics,
        )

        if not self.trace:
            return

        if self.type == "rag":
            self.trace.append_span(
                RAGSpan(
                    type=self.type,
                    name=self.name,
                    span_id=str(self.span_id),
                    parent_id=str(self.parent.span_id) if self.parent else None,
                    trace_id=str(self.trace.trace_id),
                    input=(
                        reduce_payload_size(
                            autoconvert_typed_values(self.input),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.input
                        else None
                    ),
                    output=(
                        reduce_payload_size(
                            autoconvert_typed_values(self.output),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.output
                        else None
                    ),
                    error=(
                        reduce_payload_size(
                            capture_exception(self.error),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.error
                        else None
                    ),
                    timestamps=self.timestamps,
                    contexts=reduce_payload_size(
                        autoconvert_rag_contexts(self.contexts or []),
                        max_string_length=self.trace.max_string_length,
                    ),
                    params=self.params,
                    metrics=self.metrics,
                )
            )
        elif self.type == "llm":
            self.trace.append_span(
                LLMSpan(
                    type=self.type,
                    name=self.name,
                    span_id=str(self.span_id),
                    parent_id=str(self.parent.span_id) if self.parent else None,
                    trace_id=str(self.trace.trace_id),
                    input=(
                        reduce_payload_size(
                            autoconvert_typed_values(self.input),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.input
                        else None
                    ),
                    output=(
                        reduce_payload_size(
                            autoconvert_typed_values(self.output),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.output
                        else None
                    ),
                    error=(
                        reduce_payload_size(
                            capture_exception(self.error),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.error
                        else None
                    ),
                    timestamps=self.timestamps,
                    model=self.model,
                    params=self.params,
                    metrics=self.metrics,
                )
            )
        else:
            self.trace.append_span(
                BaseSpan(
                    type=self.type,
                    name=self.name,
                    span_id=str(self.span_id),
                    parent_id=str(self.parent.span_id) if self.parent else None,
                    trace_id=str(self.trace.trace_id),
                    input=(
                        reduce_payload_size(
                            autoconvert_typed_values(self.input),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.input
                        else None
                    ),
                    output=(
                        reduce_payload_size(
                            autoconvert_typed_values(self.output),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.output
                        else None
                    ),
                    error=(
                        reduce_payload_size(
                            capture_exception(self.error),
                            max_string_length=self.trace.max_string_length,
                        )
                        if self.error
                        else None
                    ),
                    timestamps=self.timestamps,
                    params=self.params,
                    metrics=self.metrics,
                )
            )


@deprecated(
    reason="This function is deprecated, please use the @langwatch.span() decorator instead"
)
def create_span(
    name: Optional[str] = None, type: SpanTypes = "span", input: Any = None
):
    return ContextSpan(
        trace=current_trace_var.get() or None,  # type: ignore
        span_id=f"span_{nanoid.generate()}",
        name=name,
        type=type,
        input=input,
    )


@deprecated(
    reason="This function is deprecated, please use the @langwatch.rag_span() decorator instead together with langwatch.get_current_span().update(contexts=[]) to store the context"
)
def capture_rag(
    contexts: List[RAGChunk] = [],
    input: Optional[str] = None,
    name: str = "RetrievalAugmentedGeneration",
):
    return ContextSpan(
        name=name,
        type="rag",
        input={"type": "text", "value": input} if input else None,
        contexts=contexts,
    )


def get_current_trace() -> "ContextTrace":
    current_trace = current_trace_var.get()
    if not current_trace:
        raise ValueError(
            "No current trace found, could not get_current_trace(), make sure you added a @langwatch.trace() decorator on your LLM pipeline top-level function"
        )
    return current_trace


def get_current_span() -> "ContextSpan":
    current_trace = current_trace_var.get()
    if not current_trace:
        raise ValueError(
            "No current trace found, could not get_current_span(), make sure you added a @langwatch.trace() decorator on your LLM pipeline top-level function"
        )
    current_span = current_trace.get_current_span()
    if not current_span:
        raise ValueError(
            "No current span found, could not get_current_span(), make sure you added a @langwatch.span() decorator on your LLM pipeline top-level function"
        )
    return current_span


executor = ThreadPoolExecutor(max_workers=10)
current_trace_var = contextvars.ContextVar[Optional["ContextTrace"]](
    "current_trace", default=None
)


class ContextTrace:
    sent_once = False
    scheduled_send: Optional[Future[None]] = None
    _current_span = contextvars.ContextVar[Optional[ContextSpan]](
        "current_span", default=None
    )
    _current_span_global: Optional[ContextSpan] = (
        None  # Fallback for backing up edge cases like langchain
    )
    context_token: Optional[contextvars.Token[Optional["ContextTrace"]]] = None

    trace_id: Union[str, UUID]
    metadata: TraceMetadata = {}
    expected_output: Optional[str] = None
    span: Type[ContextSpan]
    root_span: Optional[ContextSpan] = None
    evaluations: List[Evaluation] = []
    disable_sending: bool = False
    max_string_length: Optional[int] = 5000
    _capture_input: bool = True
    _capture_output: bool = True
    _skip_root_span: bool = False

    api_key: Optional[str] = None
    _force_sync: bool = False

    def __init__(
        self,
        trace_id: Optional[Union[str, UUID]] = None,
        metadata: Optional[TraceMetadata] = None,
        expected_output: Optional[str] = None,
        api_key: Optional[str] = None,
        disable_sending: bool = False,
        max_string_length: Optional[int] = 5000,
        # Span constructor parameters
        span_id: Optional[str] = None,
        capture_input: bool = True,
        capture_output: bool = True,
        name: Optional[str] = None,
        type: SpanTypes = "span",
        input: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        output: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[SpanTimestamps] = None,
        contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
        model: Optional[str] = None,
        params: Optional[SpanParams] = None,
        metrics: Optional[SpanMetrics] = None,
        evaluations: Optional[List[Evaluation]] = None,
        skip_root_span: bool = False,
    ):
        self.api_key = api_key or langwatch.api_key
        self.disable_sending = disable_sending
        self.spans: Dict[str, Span] = {}
        self.max_string_length = max_string_length

        self._capture_input = capture_input
        self._capture_output = capture_output
        self._skip_root_span = skip_root_span

        self.trace_id = trace_id or f"trace_{nanoid.generate()}"
        self.metadata = metadata or {}
        self.expected_output = expected_output
        self.metadata.update({"sdk_version": get_version(), "sdk_language": "python"})
        self.evaluations = evaluations or []
        self.span = cast(
            Type[ContextSpan], lambda **kwargs: ContextSpan(trace=self, **kwargs)
        )
        if not skip_root_span:
            self.root_span = self.span(
                span_id=span_id,
                name=name,
                type=type,
                input=input,
                output=output,
                error=error,
                timestamps=timestamps,
                contexts=contexts,
                model=model,
                params=params,
                metrics=metrics,
            )

    def __enter__(self):
        get_logger().debug(f"Entered trace {self.trace_id}")
        self.context_token = current_trace_var.set(self)
        if self.root_span:
            self.root_span.__enter__()
        return self

    def __exit__(self, _type, _value, _traceback):
        get_logger().debug(f"Exiting trace {self.trace_id}")
        self.deferred_send_spans()
        if self.root_span:
            self.root_span.__exit__(_type, _value, _traceback)
        if self.context_token:
            try:
                current_trace_var.reset(self.context_token)
            except ValueError:
                pass
            self.context_token = None

    def __call__(self, func: T) -> T:
        trace_kwargs = {
            "trace_id": None,
            "metadata": self.metadata,
            "api_key": self.api_key,
            "disable_sending": self.disable_sending,
            "max_string_length": self.max_string_length,
            # Span constructor parameters
            "capture_input": self._capture_input,
            "capture_output": self._capture_output,
            "skip_root_span": self._skip_root_span,
        }

        if self.root_span:
            trace_kwargs.update(
                {
                    "name": self.root_span.name,
                    "type": self.root_span.type,
                    "input": self.root_span.input,
                    "output": self.root_span.output,
                    "error": self.root_span.error,
                    "timestamps": None,
                    "contexts": self.root_span.contexts,
                    "model": self.root_span.model,
                    "params": self.root_span.params,
                    "metrics": self.root_span.metrics,
                }
            )

        if inspect.isasyncgenfunction(func):

            @functools.wraps(func)
            async def async_gen_wrapper(*args, **kwargs):
                with ContextTrace(**trace_kwargs) as trace:
                    if trace.root_span:
                        trace.root_span._capture_name_and_input(func, *args, **kwargs)
                    items = []
                    async for item in func(*args, **kwargs):
                        items.append(item)
                        yield item

                    output = (
                        "".join(items)
                        if all(isinstance(item, str) for item in items)
                        else items
                    )
                    if trace.root_span:
                        trace.root_span._capture_output_and_maybe_name(func, output)

            return async_gen_wrapper  # type: ignore
        elif inspect.isgeneratorfunction(func):

            @functools.wraps(func)
            def sync_gen_wrapper(*args, **kwargs):
                with ContextTrace(**trace_kwargs) as trace:
                    if trace.root_span:
                        trace.root_span._capture_name_and_input(func, *args, **kwargs)
                    items = []
                    for item in func(*args, **kwargs):
                        items.append(item)
                        yield item

                    output = (
                        "".join(items)
                        if all(isinstance(item, str) for item in items)
                        else items
                    )
                    if trace.root_span:
                        trace.root_span._capture_output_and_maybe_name(func, output)

            return sync_gen_wrapper  # type: ignore
        elif inspect.iscoroutinefunction(func):

            @functools.wraps(func)
            async def async_wrapper(*args, **kwargs):
                with ContextTrace(**trace_kwargs) as trace:
                    if trace.root_span:
                        trace.root_span._capture_name_and_input(func, *args, **kwargs)
                    output = await func(*args, **kwargs)
                    if trace.root_span:
                        trace.root_span._capture_output_and_maybe_name(func, output)
                    return output

            return async_wrapper  # type: ignore
        else:

            @functools.wraps(func)  # type: ignore
            def sync_wrapper(*args, **kwargs):
                with ContextTrace(**trace_kwargs) as trace:
                    if trace.root_span:
                        trace.root_span._capture_name_and_input(func, *args, **kwargs)
                    output = func(*args, **kwargs)  # type: ignore
                    if trace.root_span:
                        trace.root_span._capture_output_and_maybe_name(func, output)
                    return output

            return sync_wrapper  # type: ignore

    def update(
        self,
        trace_id: Optional[Union[str, UUID]] = None,
        metadata: Optional[TraceMetadata] = None,
        expected_output: Optional[str] = None,
        disable_sending: Optional[bool] = None,
        # root span update
        span_id: Optional[Union[str, UUID]] = None,
        name: Optional[str] = None,
        type: Optional[SpanTypes] = None,
        input: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        output: Optional[Union[SpanInputOutput, str, List[ChatMessage]]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[SpanTimestamps] = None,
        contexts: Optional[Union[List[RAGChunk], List[str]]] = None,
        model: Optional[str] = None,
        params: Optional[SpanParams] = None,
        metrics: Optional[SpanMetrics] = None,
        evaluations: Optional[List[Evaluation]] = None,
    ):
        if trace_id:
            self.trace_id = trace_id
        if metadata:
            self.metadata.update(metadata)
        if expected_output:
            self.expected_output = expected_output
        if evaluations:
            # Avoid late mutations after capturing the value
            self.evaluations = copy.deepcopy(evaluations)

        if self.root_span:
            self.root_span.update(
                span_id=span_id,
                name=name,
                type=type,
                input=input,
                output=output,
                error=error,
                timestamps=timestamps,
                contexts=contexts,
                model=model,
                params=params,
                metrics=metrics,
            )

        if disable_sending is not None:
            self.disable_sending = disable_sending

    def add_evaluation(
        self,
        *,
        evaluation_id: Optional[str] = None,
        span: Optional[ContextSpan] = None,
        name: str,
        type: Optional[str] = None,
        is_guardrail: Optional[bool] = None,
        status: Literal["processed", "skipped", "error"] = "processed",
        passed: Optional[bool] = None,
        score: Optional[float] = None,
        label: Optional[str] = None,
        details: Optional[str] = None,
        cost: Optional[Union[Money, MoneyDict, float]] = None,
        error: Optional[Exception] = None,
        timestamps: Optional[EvaluationTimestamps] = None,
    ):
        current_evaluation_index = [
            i
            for i, e in enumerate(self.evaluations)
            if evaluation_id
            and "evaluation_id" in e
            and e["evaluation_id"] == evaluation_id
        ]
        current_evaluation_index = (
            current_evaluation_index[0] if len(current_evaluation_index) > 0 else None
        )
        current_evaluation = (
            self.evaluations[current_evaluation_index]
            if current_evaluation_index
            else None
        )

        evaluation_result = EvaluationResult(
            status=status,
        )
        if passed is not None:
            evaluation_result["passed"] = passed
        if score is not None:
            evaluation_result["score"] = score
        if label is not None:
            evaluation_result["label"] = label
        if details is not None:
            evaluation_result["details"] = details
        if cost is not None:
            if isinstance(cost, Money):
                evaluation_result["cost"] = {
                    "currency": cost.currency,
                    "amount": cost.amount,
                }
            elif isinstance(cost, float) or isinstance(cost, int):
                evaluation_result["cost"] = {"currency": "USD", "amount": cost}
            else:
                evaluation_result["cost"] = cost

        if not span:
            span = self.span(type="evaluation")
        if span.type != "evaluation":
            span = span.span(type="evaluation")
        span.update(
            name=name,
            output=TypedValueEvaluationResult(
                type="evaluation_result",
                value=evaluation_result,
            ),
            error=error,
            timestamps=(
                SpanTimestamps(
                    started_at=(
                        timestamps["started_at"]
                        if "started_at" in timestamps and timestamps["started_at"]
                        else cast(int, None)
                    ),
                    finished_at=(
                        timestamps["finished_at"]
                        if "finished_at" in timestamps and timestamps["finished_at"]
                        else cast(int, None)
                    ),
                )
                if timestamps
                else None
            ),
        )
        if "cost" in evaluation_result and evaluation_result["cost"]:
            span.update(metrics=SpanMetrics(cost=evaluation_result["cost"]["amount"]))
        span.end()

        evaluation = Evaluation(
            evaluation_id=evaluation_id or f"eval_{nanoid.generate()}",
            span_id=str(span.span_id) if span else None,
            name=name,
            type=type,
            is_guardrail=is_guardrail,
            status=status,
            passed=passed,
            score=score,
            label=label,
            details=details,
            error=capture_exception(error) if error else None,
            timestamps=timestamps,
        )

        if current_evaluation and current_evaluation_index:
            self.evaluations[current_evaluation_index] = current_evaluation | evaluation
        else:
            self.evaluations.append(evaluation)

    def evaluate(
        self,
        slug: str,
        name: Optional[str] = None,
        input: Optional[str] = None,
        output: Optional[str] = None,
        expected_output: Optional[str] = None,
        contexts: Union[List[RAGChunk], List[str]] = [],
        conversation: Conversation = [],
        settings: Optional[dict] = None,
        as_guardrail: bool = False,
    ):
        return langwatch.evaluations.evaluate(
            trace=self,
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            conversation=conversation,
            settings=settings,
            as_guardrail=as_guardrail,
        )

    async def async_evaluate(
        self,
        slug: str,
        name: Optional[str] = None,
        input: Optional[str] = None,
        output: Optional[str] = None,
        expected_output: Optional[str] = None,
        contexts: Union[List[RAGChunk], List[str]] = [],
        conversation: Conversation = [],
        settings: Optional[dict] = None,
        as_guardrail: bool = False,
    ):
        return await langwatch.evaluations.async_evaluate(
            trace=self,
            slug=slug,
            name=name,
            input=input,
            output=output,
            expected_output=expected_output,
            contexts=contexts,
            conversation=conversation,
            settings=settings,
            as_guardrail=as_guardrail,
        )

    def deferred_send_spans(self):
        get_logger().debug(
            f"Scheduling for sending trace {self.trace_id} in 1s to {langwatch.endpoint}"
        )
        self._add_finished_at_to_missing_spans()

        if "PYTEST_CURRENT_TEST" in os.environ:
            # Keep on the same thread for tests
            self.send_spans()
            return

        def run_in_thread():
            time.sleep(1)  # wait for other spans to be added
            self.sent_once = True
            self.send_spans()

        if self.scheduled_send and not self.scheduled_send.done():
            self.scheduled_send.cancel()

        self.scheduled_send = executor.submit(run_in_thread)

    def send_spans(self):
        send_spans(
            CollectorRESTParams(
                trace_id=str(self.trace_id),
                metadata=self.metadata,
                spans=list(self.spans.values()),
                expected_output=self.expected_output,
                evaluations=self.evaluations,
            ),
            api_key=self.api_key,
            force_sync=self._force_sync,
            disable_sending=self.disable_sending,
        )

    def append_span(self, span: Span):
        span["span_id"] = span.get("span_id", f"span_{nanoid.generate()}")
        self.spans[span["span_id"]] = span
        if self.sent_once:
            self.deferred_send_spans()  # send again if needed

    def get_parent_id(self):
        current_span = self.get_current_span()
        if current_span:
            return current_span.span_id
        return None

    def get_current_span(self) -> Optional[ContextSpan]:
        return self._current_span_global or self._current_span.get()

    def set_current_span(self, span: ContextSpan):
        token = self._current_span.set(span)
        self._current_span_global = span

        return token

    def reset_current_span(
        self, token: Optional[contextvars.Token[Optional["ContextSpan"]]]
    ):
        try:
            if not token:
                raise ValueError("No token provided")
            self._current_span.reset(token)
            self._current_span_global = self._current_span.get()

        # Fallback to manually set the parent back in case the span is ended in a different context and token does not work
        except ValueError:
            current_span = self.get_current_span()
            if current_span and current_span.parent:
                self.set_current_span(current_span.parent)
            else:
                self._current_span.set(None)
                self._current_span_global = None

    # Some spans may have their timestamps overwritten, never setting the finished_at, so we do it here as a fallback
    def _add_finished_at_to_missing_spans(self):
        for span in self.spans.values():
            if "timestamps" in span and (
                "finished_at" not in span["timestamps"]
                or span["timestamps"]["finished_at"] == None
            ):
                span["timestamps"]["finished_at"] = milliseconds_timestamp()

    def get_langchain_callback(self):
        import langwatch.langchain  # import dynamically here instead of top-level because users might not have langchain installed

        return langwatch.langchain.LangChainTracer(trace=self)

    def autotrack_openai_calls(
        self, client: Union["OpenAI", "AsyncOpenAI", "AzureOpenAI", "AsyncAzureOpenAI"]
    ):
        import langwatch.openai  # import dynamically here instead of top-level because, believe it or not, users might not have openai installed

        langwatch.openai.OpenAITracer(trace=self, client=client)

    def autotrack_litellm_calls(self, client: ModuleType):
        import langwatch.litellm  # import dynamically here instead of top-level because users might not have litellm installed

        langwatch.litellm.LiteLLMPatch(trace=self, client=client)

    def autotrack_dspy(self):
        langwatch.dspy.tracer(trace=self)

    def share(self):
        with httpx.Client() as client:
            response = client.post(
                f"{langwatch.endpoint}/api/trace/{self.trace_id}/share",
                headers={"X-Auth-Token": str(self.api_key)},
                timeout=15,
            )
            response.raise_for_status()
            path = response.json()["path"]
            return f"{langwatch.endpoint}{path}"

    def unshare(self):
        with httpx.Client() as client:
            response = client.post(
                f"{langwatch.endpoint}/api/trace/{self.trace_id}/unshare",
                headers={"X-Auth-Token": str(self.api_key)},
                timeout=15,
            )
            response.raise_for_status()


@retry(tries=5, delay=0.5, backoff=3)
def send_spans(
    data: CollectorRESTParams,
    api_key: Optional[str] = None,
    force_sync=False,
    disable_sending: Optional[bool] = False,
):
    import json

    get_logger().debug(
        f"Sending trace: {json.dumps(data, cls=SerializableWithStringFallback, indent=2)}"
    )

    api_key = api_key or langwatch.api_key
    if not api_key:
        warn(
            "LANGWATCH_API_KEY is not set, LLMs traces will not be sent, go to https://langwatch.ai to set it up"
        )
        return

    force_sync = f"?force_sync=true" if force_sync else ""

    if len(data["spans"]) > 256:
        warn(
            f"Going over the limit of 256 spans, dropping {len(data['spans']) - 256} spans from being sent to LangWatch"
        )
        data["spans"] = data["spans"][:256]

    # TODO: replace this with httpx, don't forget the custom SerializableWithStringFallback encoder

    if not (
        langwatch.enabled and random() < langwatch.sampling_rate and not disable_sending
    ):
        return

    with httpx.Client(timeout=60.0) as client:
        response = client.post(
            f"{langwatch.endpoint}/api/collector{force_sync}",
            json=data,
            headers={
                "X-Auth-Token": str(api_key),
                "Content-Type": "application/json",
            },
        )

        if response.status_code == 429:
            json_response = response.json()
            if (
                "message" in json_response
                and "ERR_PLAN_LIMIT" in json_response["message"]
            ):
                warn(json_response["message"])
            else:
                warn(
                    "Rate limit exceeded, dropping message from being sent to LangWatch"
                )
            return
        elif response.status_code >= 300 and response.status_code < 500:
            try:
                json_response = response.json()
                if "message" in json_response:
                    message = json_response["message"]
                    warn(f"LangWatch returned an error: {message}")
                else:
                    warn(f"LangWatch returned an error: {response.content}")
            except:
                warn(f"LangWatch returned an error: {response.content}")

        response.raise_for_status()

    get_logger().debug(f"Trace {data['trace_id']} sent successfully")


trace = ContextTrace
span = ContextSpan



from dotenv import load_dotenv

load_dotenv()
from langwatch.batch_evaluation import BatchEvaluation, DatasetEntry


def callback(entry: DatasetEntry):
    # generate messages for entry["input"] using your LLM
    # input_data = entry["input"]
    # Assuming the dataset contains an "input" column

    # Process the input data using your LLM and generate a response
    # response = f"Generated response for input: {input_data}"
    # print(response)
    # return {"output": response}

    return {"output": "Hello! How can I help you today?"}


# Instantiate the BatchEvaluation object
evaluation = BatchEvaluation(
    experiment="My Experiment",
    dataset="",  # Provide the actual dataset name here
    evaluations=[""],  # Provide the actual evaluations here
    callback=callback,
)

# Run the evaluation
results = evaluation.run()



import time
from dotenv import load_dotenv

load_dotenv()

import chainlit as cl

import langwatch


@langwatch.span(type="llm")
async def stream_tokens(tokens: list[str]):
    langwatch.get_current_span().update(model="custom_model")
    for token in tokens:
        yield token


@langwatch.trace()
async def generate(message: str):
    time.sleep(0.5)  # generating the message...

    async for token in stream_tokens(["Hello", " there! "]):
        yield token

    time.sleep(0.5)  # generating the message...

    async for token in stream_tokens(["How", " can", " I", " help?"]):
        yield token


@cl.on_message
async def main(message: cl.Message):
    msg = cl.Message(
        content="",
    )

    async for token in generate(message.content):
        await msg.stream_token(token)

    await msg.update()
